#include <ctype.h>

#include "hw.h"
#include "f020.h"

#include "bootld.h"
#include "br_setup.h"
#include "sernoi.h"
#include "print.h"

#pragma codeseg  LOADER_CODE
#pragma constseg LOADER_CONST

__bit f_valid;
__bit code_erased;

void
prompt()
{
	print("Command>");
}

void
print_menu(void)
{
	print(
		"\r\r"
		"C8051F020 Programmer\r"
   		"1. Erase and Program\r"
   		"2. Cancel\r"
	);
	prompt();
	P0= P7= 0xc0;
}

unsigned char
toint(char c)
{
	c= toupper(c);
	if (c<='9')
		return c-'0';
	else
		return c-'A'+10;
}

unsigned char
hex2char()
{
   	unsigned char retval;
   	char byteH, byteL;
   
   	// get a two-byte ASCII representation of a char from the UART
   	byteH = serial_receive();/*_getkey();*/
   	byteL = serial_receive();/*_getkey();*/
   
   	// convert to a single 8 bit result
   	retval = (char) toint(byteH) * 16;
   	retval += (char) toint(byteL);
   	return retval;
}

void
erase_flash(void)
{
   __xdata char *pagePointer = 0;	// a pointer to xdata located in data space
                                    // points to the first FLASH page that 
                                    // will be erased
      
   int i;                           // temporary int
   bit EA_state;                    // holds interrupt state

   print("\r*** Erasing 0-0x");print_x(LOADER_START);

   EA_state = EA;                   // save interrupt state

   EA = 0;                          // disable interrupts
   FLSCL |= 0x01;                   // enable FLASH write/erase
   PSCTL  = 0x03;                   // MOVX erases FLASH
   
   // Erase the first 8 FLASH pages
   for (i = 0; i < LOADER_START/512; i++){
      
      *pagePointer = 0;             // initiate the erase
      
      pagePointer += 512;           // advance to next FLASH page
	  //EA= EA_state;
	if ((i%5) == 0)
	{
	   	serial_send('.');
		//EA= 0;
		if (P0&0x80)
			P0= P7= 0x0f;
		else
			P0= P7= 0xff;
	}
   }
      
	P0= P7= 0x0f;

   PSCTL = 0x00;                    // MOVX writes target XRAM
   FLSCL &= ~0x01;                  // disable FLASH write/erase
   
   EA =  EA_state;                  // restore interrupt state

   f_valid = FALSE;                 // indicate that code is no longer valid
   code_erased = TRUE;              // indicate that FLASH has been erased
}

//-----------------------------------------------------------------------------
// receive_code
//-----------------------------------------------------------------------------
// This routine receives the new firmware through the UART in HEX record 
// format.
//
// Hex Record Format:
//
// +--------+--------+------+-------+--------+------(n bytes)------+----------+
// | RECORD | RECLEN |    OFFSET    | RECORD |                     | CHECKSUM |
// |  MARK  |  (n)   |   (2 BYTES)  |  TYPE  |        DATA         |          |
// |  ':'   |        |              |        |                     |          |
// +--------+--------+------+-------+--------+------(n bytes)------+----------+
//
//
void receive_code(void)
{
   __xdata char *pwrite;    			  // pointer used for writing FLASH
   __code char *pread;                    // pointer used for reading FLASH
   unsigned char len;                     // holds the HEX record length field
   unsigned char record_type;             // holds the HEX record type field
   unsigned int offset;                   // holds the HEX record offset field
                                          // this is the starting address of
                                          // the code image contained in the 
                                          // record

   char checksum;                         // holds the HEX record checksum field
   char flash_checksum;                   // holds the checksum calculated after
                                          // the FLASH has been programmed
   bit EA_state;                          // temporary holder used to restore 
                                          // interrupts to their previous state

   char c;                                // temporary char
   int i;                                 // temporary int
	int written= 0;

   // make sure FLASH has been erased
   if(!code_erased){
      print("\r*** FLASH not erased\r");
      return;
   } else {
      print("\rReady to receive...\r");
   }

   // wait for the user send HEX file

   do{
      
      while( c = /*_getkey()*/serial_receive() != ':' );      // ignore all characters until 
                                          // reaching the record mark field
      
      // get the record length
      len = hex2char();
      
      // get the starting address (offset field in HEX record)
      offset = hex2char();                // get the MSB
      offset <<= 8;                       
      offset |= hex2char();               // get the LSB

      // get the record type
      record_type = hex2char();
      if( record_type != 0 && record_type != 1 ){
         print("\r*** Wrong file\r"); 
         return;
      }      

      EA_state = EA;                      // save the interrupt enable bit state
      
      EA = 0;                             // disable interrupts (precautionary)   
      FLSCL |= 0x01;                      // enable FLASH write/erase
      PSCTL  = 0x01;                      // MOVX writes FLASH

      pwrite = (__xdata char *)offset;      // initialize the write pointer
      
      
      code_erased = FALSE;                // clear the code_erased flag      
           
      // write the record into flash
      for( i = 0; i < len; i++){
         
         // check for valid pointer
         if(pwrite < (__xdata char *)LOADER_START){ 
            *pwrite = hex2char();         // write one byte to FLASH
            pwrite++;                     // increment FLASH write pointer
         } else {
            print("\r\rExceeded space\r");  // print error message
         }
		 written++;
		 if ((written%256) == 0)
		 {
		 	serial_send('.');
			P0^= 0x80;
			P7^= 0x80;
		 }
      }
      P0= P7= 0x0f;
      PSCTL = 0x00;                       // MOVX writes target XRAM
      FLSCL &= ~0x01;                     // disable FLASH write/erase
      EA = EA_state;                      // restore interrupts to previous state
            
      // verify the checksum
      pread =  (__code char *)offset;       // initialize the read pointer
      checksum = hex2char();              // get the HEX record checksum field
      flash_checksum = 0;                 // set the flash_checksum to zero

      // add the data field stored in FLASH to the checksum
      for( i = 0; i < len; i++)
      {
         flash_checksum += *pread++;
      } 
      
      // add the remaining fields
      flash_checksum += len;
      flash_checksum += (char) (offset >> 8);
      flash_checksum += (char) (offset & 0x00FF);
      flash_checksum += record_type;
      flash_checksum += checksum;
      
      // verify the checksum (the flash_checksum should equal zero)     
      if(flash_checksum != 0){
         print("*** Checksum error");
         return;
      }

   } while(record_type != 1); 
   
   f_valid = TRUE;                        // indicate that download is valid 
   print("\rOK\r");
}


void
main(void)
{
	long l= 0;
	const struct t_board_info *bi;
	
	EA= 0;
	IE= 0;
	IP= 0;
	EIE1= 0;
	EIE2= 0;
	EIP1= 0;
	EIP2= 0;
	bi= get_board_info();
	silabs_init(bi?(bi->clk_table[bi->used_clk].osc):CLK_22M);
	serial_init(9600);
	br_init(bi?(bi->used_br):BR_19K);

	f_valid= 0;
	code_erased= 0;

	while ((P3 & 0xfc) != 0xfc) ;		// Wait to release all buttons
	
	print_menu();
	for (;;)
	{
		if (serial_received())
		{
			char c= serial_receive();
			switch (c)
			{
				case '1':
					erase_flash();
					receive_code();
				case '2':
					print("\r** RESETTING **\r\r");
            	    RSTSRC= 0x10;      // reset the device   
					break;
				default:
					print_menu();
					break;
			}
		}
		if ((P3 & 0x20) == 0)
		{
			br_setup();
			print_menu();
		}
		if (++l > 0x6000)
		{
			l= 0;
			P0^= 0xf0;
			P7^= 0xf0;
		}
	}
}
